#include <fstream>
#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "BitVec.hh"

BitVec::BitVec(const BitVec& bv)
{
  memcpy(this, &bv, sizeof(BitVec));

  const uint64_t num_bits = _num_trans * _item_width;
  const uint64_t num_bytes = num_bits / 8 + 4;

  if ((this->db = malloc(num_bytes)) == NULL)
  {
    perror("malloc BitVec() copy constructor");
    std::cout << "num_bytes " << num_bytes << "\n";
    exit(1);
  }

  memcpy(this->db, bv.db, num_bytes);
}

BitVec::BitVec(uint64_t _num_trans, uint64_t _min_item, uint64_t _max_item)
    : db(NULL), _num_trans(_num_trans), _min_item(_min_item), _max_item(_max_item), _item_width(_max_item - _min_item + 1)
{
  const uint64_t num_bits = _num_trans * _item_width;
  const uint64_t num_bytes = num_bits / 8 + 4;

  if ((size_t)num_bytes < num_bytes)
  {
    std::cout << "Error: won't be able to calloc " << num_bytes << " bytes on architecture with sizeof(size_t) " << sizeof(size_t) << "\n";
  }

  //std::cout << "num_trans " << num_trans << " min_item " << min_item << " max_item " << max_item << " db_size " << db_size << "\n";
  //std::cout << "BitVec allocating " << num_bits << " bits, " << num_bytes << "\n";
  if ((db = calloc(num_bytes, 1)) == NULL)
  {
    perror("calloc BitVec()");
    std::cout << "num_bytes " << num_bytes << "\n";
    exit(1);
  }
}

BitVec::~BitVec()
{
  free(db);
}

uint64_t BitVec::num_bits_set() const
{
  const size_t num_bytes = (_num_trans * _item_width) / 8 + 4;
  const uint32_t *bit_vec_end = reinterpret_cast<const uint32_t*>(db) + (num_bytes / sizeof(uint32_t));
  uint32_t *bit_vec = reinterpret_cast<uint32_t*>(db);
  //std::cout << "database covers " << bit_vec_end - bit_vec << " 32-bit regions\n";
  uint64_t num_bits = 0;
  do {
    // population count code from http://graphics.stanford.edu/~seander/bithacks.html
    uint32_t v = *bit_vec;
    v = v - ((v >> 1) & 0x55555555);                    // reuse input as temporary
    v = (v & 0x33333333) + ((v >> 2) & 0x33333333);     // temp
    num_bits += (((v + (v >> 4)) & 0xF0F0F0F) * 0x1010101) >> 24; // count
    //std::cout << "incremented num_bits to " << num_bits << "\n";
  } while (++bit_vec != bit_vec_end);

  return num_bits;
}

uint64_t BitVec::num_trans() const
{
  return this->_num_trans;
}

uint64_t BitVec::min_item() const
{
  return this->_min_item;
}

uint64_t BitVec::max_item() const
{
 return this->_max_item;
}

void BitVec::transpose()
{
  BitVec tp(_max_item - _min_item + 1, 1, _num_trans);
  //std::cout << "tranposing:\n";
  //std::cout << *this << "\n";

  for (uint64_t t=1; t<=_num_trans; ++t)
  {
    for (uint64_t i=_min_item; i<= _max_item; ++i)
    {
      if (get(t, i))
      {
        tp.set_one(i - _min_item + 1, t);
      }
    }
  }

  //std::cout << "to:\n";
  //std::cout << tp << "\n";

  _num_trans = tp._num_trans;
  _min_item = tp._min_item;
  _max_item = tp._max_item;
  _item_width = _max_item - _min_item + 1;

  std::swap(this->db, tp.db);
}
void BitVec::write_to_trans_file(const std::string& fname) const
{
  std::ofstream of(fname.c_str(), std::ios::out | std::ios::trunc | std::ios::binary);
  if (of.fail())
  {
    perror("could not open file for writing of bit vector");
    std::cerr << "Error opening file " << fname << " for writing.\n";
    exit(1);
  }
  for (std::size_t i=1; i<=_num_trans; ++i)
  {
    uint64_t items_printed = 0;
    for (std::size_t j=_min_item; j<=_max_item; ++j)
    {
      if (get(i, j) == 1)
      {
	if (++items_printed > 1)
	{
          of << " ";
	}
        of << j;
      }
    }
    of << "\n";
  }
  if (of.fail())
  {
    perror("error writing bit vector data to file");
    std::cerr << "Error writing to file " << fname << "\n";
    exit(1);
  }
  of.close();
}

std::ostream& operator<<(std::ostream& os, const BitVec& tip_db)
{
  for (std::size_t i=1; i<=tip_db._num_trans; ++i)
  {
    for (std::size_t j=tip_db._min_item; j<=tip_db._max_item; ++j)
    {
      os << tip_db.get(i, j);
    }
    os << "\n";
  }
  return os;
}
