#include <assert.h>
#include <deque>
#include <errno.h>
#include <ext/algorithm>
#include <iostream>
#include <map>
#include <set>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <tr1/unordered_map>
#include <unistd.h>
#include <vector>
#include "types.hh"

typedef int32_t item_t;
typedef int32_t trans_t;

typedef std::vector<node_t> Nodelist;
typedef std::set<node_t> Nodeset;
typedef std::map<node_t,node_t> Nodemap;
typedef std::map<node_t,Nodelist > Adjlist;
typedef std::pair<int32_t,int32_t> Edge;
typedef std::vector<Edge > Edgelist;
typedef std::deque<node_t> Nodedeque;


void bfs(const Adjlist& adjlist, Nodelist& seen_nodes_order)
{
  Nodeset seen_nodes;

  std::vector<std::pair<node_t,Nodelist> > rand_node_vec(1);
  __gnu_cxx::random_sample_n(adjlist.begin(), adjlist.end(), rand_node_vec.begin(), 1);
  node_t rand_node = rand_node_vec[0].first;

  //std::cout << "Randomly chose node " << rand_node << " to start.\n";

  Nodedeque nd;
  nd.push_back(rand_node);

  while (!nd.empty())
  {
    const node_t next_node = nd.front();
    nd.pop_front();

    //std::cout << "Popped node " << next_node << "\n";

    if (seen_nodes.find(next_node) == seen_nodes.end())
    {
      //std::cout << "  not yet seen.\n";
      Adjlist::const_iterator n = adjlist.find(next_node);
      const Nodelist& nl = n->second;
      Nodelist nl_choice(nl);
      std::random_shuffle(nl_choice.begin(), nl_choice.end());
      for (Nodelist::size_type i=0; i<nl_choice.size(); ++i)
      {
        //std::cout << "  Considering neighbor " << nl[i] << "\n";
        if (seen_nodes.find(nl_choice[i]) == seen_nodes.end())
	{
	  //std::cout << "    neighbor " << nl[i] << " not yet seen; adding to queue.\n";
          nd.push_back(nl_choice[i]);
	}
      }
      seen_nodes.insert(next_node);
      seen_nodes_order.push_back(next_node);
    }
  }
}

void mcs(const Adjlist& adjlist, Nodelist& seen_nodes_order)
{
  Nodeset seen_nodes;

  std::vector<std::pair<node_t,Nodelist> > rand_node_vec(1);
  __gnu_cxx::random_sample_n(adjlist.begin(), adjlist.end(), rand_node_vec.begin(), 1);
  node_t rand_node = rand_node_vec[0].first;

  Nodedeque nd;
  nd.push_back(rand_node);

  while (!nd.empty())
  {
    const node_t next_node = nd.front();
    nd.pop_front();
    if (seen_nodes.find(next_node) == seen_nodes.end())
    {
      Adjlist::const_iterator n = adjlist.find(next_node);
      const Nodelist& nl = n->second;
      Nodelist nl_choice(nl);
      std::multimap<int,node_t> num_com_to_node;
      const Nodelist& nn_neigh = adjlist.find(next_node)->second;
      std::cout << "considering unseen node " << next_node << " with neighbors ";
      std::copy(nn_neigh.begin(), nn_neigh.end(), std::ostream_iterator<node_t>(std::cout,", "));
      std::cout << "\n";
      for (std::vector<int>::size_type i=0; i<nl_choice.size(); ++i)
      {
	const node_t cur_node = nl_choice[i];
	if (seen_nodes.find(cur_node) == seen_nodes.end())
	{
	  const Nodelist& curn_neigh = adjlist.find(cur_node)->second;
	  Nodelist res;
	  std::back_insert_iterator<Nodelist > res_ins(res);
	  std::set_intersection(nn_neigh.begin(), nn_neigh.end(), curn_neigh.begin(), curn_neigh.end(), res_ins);
	  num_com_to_node.insert(std::make_pair(res.size(), nl_choice.at(i)));
	}
      }
      for (std::multimap<int,int>::const_iterator ciIt = num_com_to_node.begin(); ciIt != num_com_to_node.end(); ++ciIt)
      {
        std::cout << ciIt->first << " " << ciIt->second << "\n";
	nd.push_back(ciIt->second);
      }
      seen_nodes.insert(next_node);
      seen_nodes_order.push_back(next_node);
    }
  }
}

void populate_adj_list(Adjlist& adjlist, std::string fname, bool include_trans, bool include_items)
{
  assert(adjlist.empty());
  assert(include_trans || include_items);

  FILE *d;
  if ((d = fopen(fname.c_str(), "rb")) == NULL)
  {
    perror("fopen");
    exit(2);
  }
  
  char *line = NULL;
  size_t line_size = 0;
  trans_t trans_num = 0;
  while (getline(&line, &line_size, d) != -1)
  {
    ++trans_num;

    item_t item;
    int item_str_len;
    char *line_scan = line;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      int trans_node_label = trans_num;
      if (include_trans && include_items)
      {
        trans_node_label *= -1;
      }

      if (include_trans)
      {
        adjlist[trans_node_label].push_back(item);
      }
      if (include_items)
      {
        adjlist[item].push_back(trans_node_label);
      }
      line_scan += item_str_len;
    }
  }
  if (line != NULL)
  {
    free(line);
  }

  if (fclose(d) != 0)
  {
    perror("fclose");
    exit(3);
  }
}

int main(int argc, char **argv)
{
  if (argc < 2)
  {
    std::cerr << "Usage: " << argv[0] << " dataset\n";
    exit(1);
  }

  struct timeval tv;
  gettimeofday(&tv, 0);
  srand48(tv.tv_sec + tv.tv_usec);
  srandom(tv.tv_sec + tv.tv_usec);

  for (int arg_i=1; arg_i < argc; ++arg_i)
  {
    std::string input_fname(argv[arg_i]);
  
    Adjlist adjlist;

    populate_adj_list(adjlist, input_fname, true, true);

    std::cout << "Adjacency list contains " << adjlist.size() << " nodes.\n";
  
    std::string input_f_basename = input_fname.substr(9);
    input_f_basename.resize(input_f_basename.find("."));
   
    Nodelist seen_nodes_order;
  
    bfs(adjlist, seen_nodes_order);
    //mcs(adjlist, seen_nodes_order);
  
    /*
    std::cout << "seen_nodes_order:\n";
    std::copy(seen_nodes_order.begin(), seen_nodes_order.end(), std::ostream_iterator<node_t>(std::cout, ", "));
    std::cout << "\n";
    */
  
    Nodemap nodemap;
    int seen_trans = 0;
    int seen_items = 0;
  
    for (Nodelist::size_type i=0; i<seen_nodes_order.size(); ++i)
    {
      node_t node = seen_nodes_order[i];
      if (node < 0)
      {
        nodemap[node] = seen_trans;
        ++seen_trans;
      } else if (node > 0) {
        nodemap[node] = seen_items;
        ++seen_items;
      }
    }
  
    {
      Adjlist orig_adjlist;
      for (Adjlist::const_iterator aIt=adjlist.begin(); aIt != adjlist.end(); ++aIt)
      {
        const node_t node = aIt->first;
        if (node < 0)
        {
          orig_adjlist[node * -1] = aIt->second;
        }
      }
  
      std::string orig_out_fname("results/" + input_f_basename + "_orig.png");
  
      write_image(orig_adjlist, orig_out_fname.c_str(), seen_items, seen_trans);
    }
  
  
    Adjlist trans_adjlist;
    for (Adjlist::const_iterator aIt=adjlist.begin(); aIt != adjlist.end(); ++aIt)
    {
      const node_t node = aIt->first;
      if (node < 0)
      {
        Nodelist nl(aIt->second);
        for (Nodelist::size_type i=0; i<nl.size(); ++i)
        {
          nl[i] = nodemap[nl[i]];
        }
        trans_adjlist[nodemap[aIt->first]] = nl;
      }
    }
  
    std::string out_filename("results/" + input_f_basename + "_bfs.png");
  
    write_image(trans_adjlist, out_filename.c_str(), seen_items, seen_trans);
  
  }

  return 0;
}
